# Copyright (c) 2013-2022 khal contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

try:
    from khal.version import version
except ImportError:
    version = "invalid"
    import sys

    sys.exit(
        "Failed to find (autogenerated) version.py. This might be due to "
        "using GitHub's tarballs or svn access. Either clone "
        "from GitHub via git or get a tarball from PyPI."
    )

__productname__ = "khal"
__version__ = version
__author__ = "Christian Geier"
__copyright__ = "Copyright (c) 2013-2022 khal contributors"
__author_email__ = "khal@lostpackets.de"
__description__ = "A standards based terminal calendar"
__license__ = "Expat/MIT, see COPYING"
__homepage__ = "https://lostpackets.de/khal/"
